import os
import torch

from transformers import AutoTokenizer, TrainingArguments, AutoModelForCausalLM, BitsAndBytesConfig
from datasets import load_dataset
from trl import DPOTrainer
from peft import LoraConfig, prepare_model_for_kbit_training, get_peft_model
from utils import find_all_linear_names, print_trainable_parameters

output_dir="./dpo_results"
model_name = "TinyLlama/TinyLlama-1.1B-intermediate-step-1431k-3T"

#dataset = load_dataset("json", data_files="dpo_conversations.json",split="train")


bnb_config = BitsAndBytesConfig(
    load_in_4bit=True,
    bnb_4bit_quant_type="nf4",
    bnb_4bit_compute_dtype=torch.bfloat16,
)

print("1")
model = AutoModelForCausalLM.from_pretrained(model_name, torch_dtype=torch.bfloat16, use_flash_attention_2=True, quantization_config=bnb_config)
model.config.use_cache = False
model = prepare_model_for_kbit_training(model)

model_ref = AutoModelForCausalLM.from_pretrained(model_name, torch_dtype=torch.bfloat16, use_flash_attention_2=True, quantization_config=bnb_config)

tokenizer = AutoTokenizer.from_pretrained(model_name)
tokenizer.pad_token = tokenizer.eos_token



def return_prompt_and_responses(sample):
    #print(sample["prompt"])
    return {
        "prompt": f"<|im_start|>system\n{sample['system']}\n<|im_end|>\n<|im_start|>user\n{sample['prompt']}\n<|im_end|>\n<|im_start|>assistant\n", #samples["prompt"],
        "chosen": sample["chosen"],
        "rejected": sample["rejected"],
    }

dataset = load_dataset("jondurbin/truthy-dpo-v0.1")#
original_columns = dataset.column_names

dataset = dataset.map(
    return_prompt_and_responses,
    #num_proc = 8,
    #batched=True,
    #remove_columns=original_columns
)
dataset.save_to_disk("dpo_dataset")


#from datasets import load_from_disk

#dataset = load_from_disk("dpo_dataset")

print("2")

training_args = TrainingArguments(
    per_device_train_batch_size=2,
    gradient_accumulation_steps=2,
    gradient_checkpointing =True,
    max_grad_norm= 0.3,
    num_train_epochs=8, 
    save_steps= 1000,
    learning_rate=2e-4,
    bf16=True,
    save_total_limit=3,
    logging_steps=10,
    output_dir=output_dir,
    optim="paged_adamw_32bit",
    lr_scheduler_type="cosine",
    warmup_ratio=0.05,
    remove_unused_columns=False
)

peft_config = LoraConfig(
    r=32,
    lora_alpha=16,
    target_modules=find_all_linear_names(model),
    lora_dropout=0.05,
    bias="none",
    task_type="CAUSAL_LM",
)
model = get_peft_model(model, peft_config)
print_trainable_parameters(model)

dpo_trainer = DPOTrainer(
    model,
    model_ref,
    args=training_args,
    beta=0.1,
    train_dataset=dataset['train'],
    tokenizer=tokenizer,
    max_prompt_length=2048,
    max_length=4096,
)


dpo_trainer.train()
dpo_trainer.save_model(output_dir)


output_dir = os.path.join(output_dir, "final_checkpoint")
dpo_trainer.model.save_pretrained(output_dir)
tokenizer.save_pretrained(output_dir)